/******************************************************************************
**
**  Module Name:    SeqTest.c
**
**  Description:    Performs sequential read tests on SCSI disks and CD-ROMs.
**                  The code demonstrates the basic use of a number of ASPI
**                  features including large buffer transfers, Plug'n'Play,
**                  forced bus rescan, timeouts, and the proper use of event
**                  notification.
**
**  Entry Points:   WinMain
**
**  History:        11/30/96 (dap)  Initial creation.
**
**  Notes:          This file created using 4 spaces per tab.
**
**  Copyright 1996-1997 Adaptec, Inc.,  All Rights Reserved.
**
******************************************************************************/

#define STRICT
#include <windows.h>
#include <string.h>
#include <commctrl.h>
#include <ctl3d.h>
#include <dbt.h>

#include "wnaspi32.h"
#include "scsidefs.h"
#include "resource.h"
#include "seqtest.h"

/*
*******************************************************************************
** Globals
*******************************************************************************
*/

HWND            ghwndDlg;                       // Handle to our dialog box
HINSTANCE       ghinstSeqTest;                  // Application instance

char            *gszAppName = "SEQTEST";        // Short name for application
char            gszFullAppName[256];            // Full file name

char            gszMsg[256];                    // Temporary message strings
char            gszMsg2[256];

BOOL            gbOldShell;                     // TRUE if we need CTL3D
HINSTANCE       ghinstCTL3D32;                  // Handle to CTL3D DLL
HINSTANCE       ghinstWNASPI32;                 // Handle to ASPI for Win32

BOOL            gbASPIBuffer;                   // TRUE if ASPI transfer buffer
ASPI32BUFF      gab;                            // Pointer to transfer buffer

BOOL            gbRefreshNeeded;                // True if new device list.
HANDLE          ghsemBenchTargets;              // Target list semaphore.
PBENCHTARGET    gpbtHead;                       // Head of the target list.

DWORD           (*gpfnGetASPI32SupportInfo)( VOID );
DWORD           (*gpfnSendASPI32Command)( LPSRB );
BOOL            (*gpfnGetASPI32Buffer)( PASPI32BUFF );
BOOL            (*gpfnFreeASPI32Buffer)( PASPI32BUFF );
BOOL            (*gpfnTranslateASPI32Address)( PDWORD, PDWORD );

BOOL            (WINAPI *gpfnCtl3dRegister)( HINSTANCE );
BOOL            (WINAPI *gpfnCtl3dUnregister)( HINSTANCE );
BOOL            (WINAPI *gpfnCtl3dAutoSubclass)( HINSTANCE );
BOOL            (WINAPI *gpfnCtl3dSubclassDlgEx)( HWND, DWORD );
BOOL            (WINAPI *gpfnCtl3dColorChange)( VOID );

/*
*******************************************************************************
** Prototypes
*******************************************************************************
*/

BOOL InitInstance( VOID );
VOID TermInstance( VOID );
BOOL InitASPI( VOID );
VOID TermASPI( VOID );

VOID InitDialog( int );
LRESULT APIENTRY SeqTestWndProc( HWND, UINT, WPARAM, LPARAM );
VOID OnTimer( VOID );
VOID OnDeviceChange( WPARAM, LPARAM );
VOID OnSysColorChange( WPARAM, LPARAM );
VOID OnCommand( WPARAM, LPARAM );
VOID OnCommandTargetList( WPARAM, LPARAM );
VOID OnCommandTest( VOID );
VOID OnCommandRescan( VOID );
VOID OnCommandAbout( VOID );
VOID OnCommandExit( VOID );

VOID RefreshTargetList( VOID );
DWORD BenchmarkThread( PBENCHTARGET );
BOOL SyncExecWithRetry( PBENCHTARGET, BYTE, DWORD, PBYTE, BYTE, PBYTE );
BOOL ClaimBenchTargetsSemaphore( VOID );
VOID ReleaseBenchTargetsSemaphore( VOID );

VOID DisplayWNASPI32Failure( DWORD, BYTE );
BOOL CALLBACK SCSIFailureDlgProc( HWND, UINT, WPARAM, LPARAM );
BOOL CALLBACK AboutDlgProc( HWND, UINT, WPARAM, LPARAM );
LPSTR GetString( UINT );

//*****************************************************************************
//*****************************************************************************

int WINAPI WinMain
(
    HINSTANCE       hInstance,
    HINSTANCE       hPrevInstance,
    LPSTR           lpszCmdLine,
    int             nCmdShow
)
{
    MSG             msg;
    HWND            hwndPrevInstance;
    OSVERSIONINFO   osvi;

    /*
    ** Get the full application title for SeqTest.  We need this for all of
    ** our dialog boxes and other error messages.  Quit if we can't get the
    ** name because something must be really wrong with resources.
    */

    ghinstSeqTest = hInstance;
    if ( !LoadString( ghinstSeqTest, IDS_APPNAME, gszFullAppName, sizeof(gszFullAppName) ) )
    {
        return 0;
    }

    /*
    ** Make sure we're not running under Win32s.  ASPI for Win32 does not
    ** run under Win32s so we'll just jump ship if we detect it.  We'll also
    ** take advantage of our access to OSVERSIONINFO to note if we need CTL3D
    ** and whether this is NT or not.
    */

    osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    GetVersionEx( &osvi );
    if( osvi.dwPlatformId == VER_PLATFORM_WIN32s )
    {
        if( LoadString( ghinstSeqTest, IDS_WIN32S, gszMsg, sizeof(gszMsg) ) )
        {
            MessageBox( NULL, gszMsg, gszFullAppName, MB_ICONSTOP | MB_OK );
        }
        return 0;
    }
    gbOldShell = (osvi.dwMajorVersion < 4) ? TRUE : FALSE;

    /*
    ** If another copy of SeqTest is running, switch to it.
    */

    hwndPrevInstance = FindWindow( gszAppName, NULL );
    if( hwndPrevInstance )
    {
        if( IsIconic( hwndPrevInstance ) )
        {
            ShowWindow( hwndPrevInstance, SW_RESTORE );
        }
        SetForegroundWindow( hwndPrevInstance );
        return 0;
    }

    /*
    ** Initialize this sucker, show the dialog, and then pump up the messages!
    ** If anything goes wrong on the init of the app (which includes creation
    ** of the dialog) then we'll just bail out.  Else we'll run till exit!
    */
    
    msg.wParam = 0;
    if( InitInstance() )
    {
        InitDialog( nCmdShow );
        while( GetMessage( &msg, NULL, 0, 0 ) )
        {
            if ( !IsDialogMessage( ghwndDlg, &msg ) )
            {
                TranslateMessage( &msg );
                DispatchMessage( &msg );
            }
        }

    }
    TermInstance();

    return msg.wParam;
}

//*****************************************************************************
//*****************************************************************************

BOOL InitInstance( VOID )
{
    WNDCLASS            wc;

    /*
    ** Initialize ASPI and allocate a transfer buffer.
    */

    if( !InitASPI() )
    {
        return FALSE;
    }
     
    /*
    ** Crank up 3D controls and let it auto-subclass everything.  We'll only
    ** try to do this if 3D controls aren't built into the system (i.e. the
    ** shell version is < 4).
    */

    if( gbOldShell )
    {
        ghinstCTL3D32 = LoadLibrary( "CTL3D32" );
        if( ghinstCTL3D32 )
        {
            (FARPROC)gpfnCtl3dRegister = GetProcAddress( ghinstCTL3D32, "Ctl3dRegister" );
            (FARPROC)gpfnCtl3dUnregister = GetProcAddress( ghinstCTL3D32, "Ctl3dUnregister" );
            (FARPROC)gpfnCtl3dAutoSubclass = GetProcAddress( ghinstCTL3D32, "Ctl3dAutoSubclass" );
            (FARPROC)gpfnCtl3dSubclassDlgEx = GetProcAddress( ghinstCTL3D32, "Ctl3dSubclassDlgEx" );
            (FARPROC)gpfnCtl3dColorChange = GetProcAddress( ghinstCTL3D32, "Ctl3dColorChange" );

            if( !gpfnCtl3dRegister     || !gpfnCtl3dUnregister    || 
                !gpfnCtl3dAutoSubclass || !gpfnCtl3dSubclassDlgEx ||
                !gpfnCtl3dColorChange )
            {
                FreeLibrary( ghinstCTL3D32 );
                ghinstCTL3D32 = NULL;                
            }
            else
            {
                gpfnCtl3dRegister( ghinstSeqTest );
                gpfnCtl3dAutoSubclass( ghinstSeqTest );
            }
        }
    }
    
    /*
    ** Initialize the common control DLL.  We use one of the new progress bars
    ** in this little application so we need this call.
    */

    InitCommonControls();
        
    /*
    ** Register a window class for our main dialog.  If the call fails quit.
    */

    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = SeqTestWndProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = DLGWINDOWEXTRA;
    wc.hInstance = ghinstSeqTest;
    wc.hIcon = LoadIcon( ghinstSeqTest, MAKEINTRESOURCE(IDR_MAINFRAME) );
    wc.hCursor = LoadCursor( NULL, IDC_ARROW );
    wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE + 1);
    wc.lpszMenuName = NULL;
    wc.lpszClassName = gszAppName;
 
    if( !RegisterClass( &wc ) )
    {
        return FALSE;
    }

    /*
    ** Create the dialog itself and then let the Ctl3D stuff subclass it if 
    ** we need to.  If the creation fails then fail.
    */

    ghwndDlg = CreateDialog( ghinstSeqTest, MAKEINTRESOURCE(IDD_SEQTEST_DIALOG), 0, NULL );
    if( ghwndDlg )
    {
        if( ghinstCTL3D32 )
        {
            gpfnCtl3dSubclassDlgEx( ghwndDlg, CTL3D_ALL );
        }
    }
    else
    {
        return FALSE;
    }

    /*
    ** Everything worked so return success!
    */

    return TRUE;
}

//*****************************************************************************
//*****************************************************************************

VOID TermInstance( VOID )
{
    /*
    ** Close down ASPI and release our transfer buffer.
    */

    TermASPI();

    /*
    ** Close down CTL3D32 if it was loaded.
    */

    if( ghinstCTL3D32 )
    {
        gpfnCtl3dUnregister( ghinstSeqTest );
        FreeLibrary( ghinstCTL3D32 );
    }
}

//*****************************************************************************
//*****************************************************************************

BOOL InitASPI( VOID )
{
    UINT                fuPrevErrorMode;
    DWORD               dwSupportInfo;
    DWORD               dwLastError;
    DWORD               dwMaxTransfer;
    SRB_GetSetTimeouts  srbTimeouts;

    /*
    ** Load ASPI for Win32.  If it won't load display a warning.
    */

    fuPrevErrorMode = SetErrorMode( SEM_NOOPENFILEERRORBOX );
    ghinstWNASPI32 = LoadLibrary( "WNASPI32" );
    dwLastError = GetLastError();
    SetErrorMode( fuPrevErrorMode );

    if( !ghinstWNASPI32 )
    {
        DisplayWNASPI32Failure( dwLastError, SS_COMP );
        return FALSE;
    }

    /*
    ** Get the ASPI entry points.  Note that only two functions are mandatory:
    ** GetASPI32SupportInfo and SendASPI32Command.  The code will run if the
    ** others are not present.
    */

    (FARPROC)gpfnGetASPI32SupportInfo = GetProcAddress( ghinstWNASPI32, "GetASPI32SupportInfo" );
    (FARPROC)gpfnSendASPI32Command = GetProcAddress( ghinstWNASPI32, "SendASPI32Command" );
    (FARPROC)gpfnGetASPI32Buffer = GetProcAddress( ghinstWNASPI32, "GetASPI32Buffer" );
    (FARPROC)gpfnFreeASPI32Buffer = GetProcAddress( ghinstWNASPI32, "FreeASPI32Buffer" );
    (FARPROC)gpfnTranslateASPI32Address = GetProcAddress( ghinstWNASPI32, "TranslateASPI32Address" );

    if( !gpfnGetASPI32SupportInfo || !gpfnSendASPI32Command )
    {
        DisplayWNASPI32Failure( ERROR_DLL_NOT_FOUND, SS_COMP );
        return FALSE;
    }

    /*
    ** Check if ASPI initialized ok.  Display an error box if it didn't.
    */

    dwSupportInfo = gpfnGetASPI32SupportInfo();
    if( HIBYTE(LOWORD(dwSupportInfo)) != SS_COMP &&
        HIBYTE(LOWORD(dwSupportInfo)) != SS_NO_ADAPTERS )
    {
        DisplayWNASPI32Failure( ERROR_SUCCESS, HIBYTE(LOWORD(dwSupportInfo)) );
        return FALSE;
    }

    /*
    ** Set timeouts for ALL devices to 15 seconds.  Nothing we deal with should
    ** take that long to do ANYTHING.  We are just doing inquiries to most
    ** devices, and then simple reads to CDs, disks, etc. so 10 seconds (even
    ** if they have to spin up) should be plenty. 
    */

    memset( &srbTimeouts, 0, sizeof(SRB_GetSetTimeouts) );
    srbTimeouts.SRB_Cmd = SC_GETSET_TIMEOUTS;
    srbTimeouts.SRB_HaId = 0xFF;
    srbTimeouts.SRB_Flags = SRB_DIR_OUT;
    srbTimeouts.SRB_Target = 0xFF;
    srbTimeouts.SRB_Lun = 0xFF;
    srbTimeouts.SRB_Timeout = 15*2;
    gpfnSendASPI32Command( &srbTimeouts );

    /*
    ** Allocate as big a transfer buffer (up to 512KB) as we can.  We do this
    ** by trying to allocate 512, 256, or 128 as big ASPI buffers.  If this
    ** all fails then we switch to VirtualAlloc and grab 64KB.  If this fails
    ** then we fail the initialization.
    */

    gbASPIBuffer = gpfnGetASPI32Buffer ? TRUE : FALSE;

    if( gbASPIBuffer )
    {
        gbASPIBuffer = FALSE;
        for( dwMaxTransfer = 524288lu; dwMaxTransfer > 65536lu; dwMaxTransfer >>= 1 )
        {
            memset( &gab, 0, sizeof(ASPI32BUFF) );
            gab.AB_BufLen = dwMaxTransfer;
            gab.AB_ZeroFill = FALSE;

            gbASPIBuffer = gpfnGetASPI32Buffer( &gab );
            if( gbASPIBuffer )
            {
                break;
            }
        }
    }

    if( !gbASPIBuffer )
    {
        gab.AB_BufLen = 65536lu;
        gab.AB_BufPointer = (PBYTE)VirtualAlloc
        (
            NULL,
            gab.AB_BufLen,
            MEM_COMMIT,
            PAGE_READWRITE
        );
    }

    if( !gab.AB_BufPointer )
    {
        DisplayWNASPI32Failure( ERROR_SUCCESS, SS_INSUFFICIENT_RESOURCES );
        return FALSE;
    }

    return TRUE;
}

//*****************************************************************************
//*****************************************************************************

VOID TermASPI( VOID )
{
    /*
    ** Free our transfer buffer.  If it is an ASPI32BUFF then use ASPI to free
    ** the buffer, else just do a VirtualFree.
    */

    if( gab.AB_BufPointer )
    {
        if( gbASPIBuffer )
        {
            gpfnFreeASPI32Buffer( &gab );
        }
        else
        {
            VirtualFree( gab.AB_BufPointer, 0, MEM_RELEASE );        
        }
    }

    /*
    ** Unload ASPI if it has been loaded.
    */

    if( ghinstWNASPI32 )
    {
        FreeLibrary( ghinstWNASPI32 );
    }
}

//*****************************************************************************
//*****************************************************************************

VOID InitDialog( int nCmdShow )
{
    int     x;
    int     y;
    int     wDlg;
    int     hDlg;
    int     wScreen;
    int     hScreen;
    RECT    rc;

    /*
    ** Center the dialog horizontally and put it 1/3 of the way down vertically.
    */

    GetWindowRect( ghwndDlg, &rc );
    wScreen = GetSystemMetrics( SM_CXSCREEN );
    hScreen = GetSystemMetrics( SM_CYSCREEN );

    x = (wScreen - (rc.right - rc.left)) / 2;
    y = (hScreen - (rc.bottom - rc.top)) / 3;
    wDlg  = rc.right - rc.left;
    hDlg = rc.bottom - rc.top;

    MoveWindow( ghwndDlg, x, y, wDlg, hDlg, 0 );

    /*
    ** Force rescan of the SCSI bus.  This will setup the list box
    ** and enable/disable the different controls as needed.
    */
    
    RefreshTargetList();

    /*
    ** Set our periodic timer and display the dialog.
    */

    SetTimer( ghwndDlg, 0, 1000, NULL );
    ShowWindow( ghwndDlg, nCmdShow );
}

//*****************************************************************************
//*****************************************************************************

LRESULT APIENTRY SeqTestWndProc
(
    HWND                    hwndDlg,
    UINT                    uMsg,
    WPARAM                  wParam,
    LPARAM                  lParam
)
{
    switch( uMsg )
    {
        case WM_TIMER:
            OnTimer();
            return 0;

        case WM_DEVICECHANGE:
            OnDeviceChange( wParam, lParam );
            return TRUE;

        case WM_COMMAND:
            OnCommand( wParam, lParam );
            return 0;
            
        case WM_SYSCOLORCHANGE:
            OnSysColorChange( wParam, lParam );
            break;

        case WM_DESTROY:
            PostQuitMessage(0);
            return 0;
    }

    return DefWindowProc( hwndDlg, uMsg, wParam, lParam );
}

//*****************************************************************************
//*****************************************************************************

VOID OnTimer( VOID )
{
    if( gbRefreshNeeded )
    {
        RefreshTargetList();
    }
}
   
//*****************************************************************************
//*****************************************************************************

VOID OnDeviceChange( WPARAM wParam, LPARAM lParam )
{
    DWORD                   dwDEVNODE;
    DWORD                   dwTargetPath;
    PDEV_BROADCAST_HDR      pHeader;
    PDEV_BROADCAST_DEVNODE  pDevnodeData;
    
    /*
    ** We only care about arrival and removal, and then only if
    ** the ASPI translation function is available.
    */

    if( !gpfnTranslateASPI32Address || 
        (wParam != DBT_DEVICEARRIVAL && wParam != DBT_DEVICEREMOVECOMPLETE) )
    {
        return;
    }

    pHeader = (PDEV_BROADCAST_HDR)lParam;
    pDevnodeData = (PDEV_BROADCAST_DEVNODE)lParam;
    if( pHeader->dbch_devicetype != DBT_DEVTYP_DEVNODE )
    {
        return;
    }

    dwDEVNODE = pDevnodeData->dbcd_devnode;
    if( !gpfnTranslateASPI32Address( &dwTargetPath, &dwDEVNODE ) ||
        ((dwTargetPath & 0xFFFFlu) == 0xFFFFlu) )
    {
        return;
    }

    /*
    ** We have an event dircly related to an ASPI SCSI target.  Set
    ** the flag which forces a benchmark target list refresh.
    */

    gbRefreshNeeded = TRUE;
}

//*****************************************************************************
//*****************************************************************************

VOID OnSysColorChange( WPARAM wParam, LPARAM lParam )
{
    if( ghinstCTL3D32 )
    {
        gpfnCtl3dColorChange();
    }
    SendMessage( GetDlgItem( ghwndDlg, IDC_PROGRESS ), WM_SYSCOLORCHANGE, wParam, lParam );
}
    
//*****************************************************************************
//*****************************************************************************

VOID OnCommand( WPARAM wParam, LPARAM lParam )
{
    switch( LOWORD(wParam) )
    {
        case IDC_TARGETLIST:
            OnCommandTargetList( wParam, lParam );
            break;

        case IDC_TEST:
            OnCommandTest();
            break;

        case IDC_RESCAN:
            OnCommandRescan();
            break;
 
        case IDC_ABOUT:
            OnCommandAbout();
            break;

        case IDOK:
            OnCommandExit();
            break;            
    }
}

//*****************************************************************************
//*****************************************************************************

VOID OnCommandTargetList( WPARAM wParam, LPARAM lParam )
{
    int             i;
    BOOL            bEnable;
    WPARAM          iTarget;
    PBENCHTARGET    pbt;

    /*
    ** Only work with selection changes.
    */

    if( HIWORD(wParam) != LBN_SELCHANGE )
    {
        return;        
    }

    /*
    ** Find the current selection.  If we can' then display all ticks.
    */

    iTarget = (WPARAM)SendDlgItemMessage( ghwndDlg, IDC_TARGETLIST, LB_GETCURSEL, 0, 0 );
    if( iTarget == LB_ERR )
    {
        for( i = 0; i < SAMPLES; i++ )
        {
            bEnable = !(((4lu << i) * 1024lu) > gab.AB_BufLen);
            EnableWindow( GetDlgItem( ghwndDlg, IDC_SIZE4 + i ), bEnable );
            EnableWindow( GetDlgItem( ghwndDlg, IDC_RATE4 + i ), bEnable );
            SetDlgItemText( ghwndDlg, IDC_RATE4 + i, "--" );
        }
        return;
    }

    /*
    ** There is a valid selection so display its statistics.
    */

    pbt = (PBENCHTARGET)SendDlgItemMessage
    (
        ghwndDlg,
        IDC_TARGETLIST,
        LB_GETITEMDATA,
        iTarget,
        0
    );

    for( i = 0; i < SAMPLES; i++ )
    {
        if( pbt->dwThroughput[i] != 0xFFFFFFFF )
        {
            wsprintf( gszMsg, "%ld", pbt->dwThroughput[i] );
        }
        else
        {
            strcpy( gszMsg, "--" );
        }

        bEnable = !(((4lu << i) * 1024lu) > pbt->dwMaxTransferBytes);
        EnableWindow( GetDlgItem( ghwndDlg, IDC_SIZE4 + i ), bEnable );
        EnableWindow( GetDlgItem( ghwndDlg, IDC_RATE4 + i ), bEnable );
        SetDlgItemText( ghwndDlg, IDC_RATE4 + i, gszMsg );
    }
}

//*****************************************************************************
//*****************************************************************************

VOID OnCommandTest( VOID )
{
    DWORD           dwThreadId;
    HANDLE          hthread;
    WPARAM          iTarget;
    PBENCHTARGET    pbt;
    
    /*
    ** Figure out who is about to be benchmarked.  If nothing is selected then
    ** just get out and let the user try again.
    */
                                       
    iTarget = (WPARAM)SendDlgItemMessage( ghwndDlg, IDC_TARGETLIST, LB_GETCURSEL, 0, 0 );
    if( iTarget == LB_ERR )
    {
        return;
    }

    pbt = (PBENCHTARGET)SendDlgItemMessage
    (
        ghwndDlg,
        IDC_TARGETLIST,
        LB_GETITEMDATA,
        iTarget,
        0
    );

    /*
    ** Try to claim the benchmark semaphore.  Doing so will make sure that
    ** we are safe to use ASPI before we go off and create any threads.  Once
    ** claimed it will remain "ours" until the benchmark thread about to be
    ** created releases it.  If we can't get the semaphore then the user
    ** will have to try again.
    */

    if( ClaimBenchTargetsSemaphore() )
    {
        /*
        ** Start the benchmark and get out.
        */

        hthread = CreateThread
        (
            NULL,
            0,
            (LPTHREAD_START_ROUTINE)BenchmarkThread,
            pbt,
            0,
            &dwThreadId
        );
        CloseHandle( hthread );
    }
}    

//*****************************************************************************
//*****************************************************************************

VOID OnCommandRescan( VOID )
{
    BYTE                    byHaId;
    BYTE                    byMaxHaId;
    DWORD                   dwASPIStatus;
    SRB_RescanPort          srbRescanPort;

    if( ClaimBenchTargetsSemaphore() )
    {
        dwASPIStatus = gpfnGetASPI32SupportInfo();
        if( HIBYTE(LOWORD(dwASPIStatus)) == SS_COMP )
        {
            byMaxHaId = LOBYTE(LOWORD(dwASPIStatus));
            for( byHaId = 0; byHaId < byMaxHaId; byHaId++ )
            {
                memset( &srbRescanPort, 0, sizeof(SRB_RescanPort) );
                srbRescanPort.SRB_Cmd = SC_RESCAN_SCSI_BUS;
                srbRescanPort.SRB_HaId = byHaId;
                gpfnSendASPI32Command( (LPSRB)&srbRescanPort );
            }
        }
        ReleaseBenchTargetsSemaphore();
    }
}

//*****************************************************************************
//*****************************************************************************

VOID OnCommandAbout( VOID )
{
    DialogBox
    (
        ghinstSeqTest,
        MAKEINTRESOURCE(IDD_ABOUTBOX),
        ghwndDlg,
        AboutDlgProc
    );
}

//*****************************************************************************
//*****************************************************************************

VOID OnCommandExit( VOID )
{
    /*
    ** Here we'll claim the benchmark semaphore just like everyone else.  If
    ** we can't get it then don't exit.  Note that an INFINITE wait for the
    ** semaphore here is unwise because it shuts down the message loop.
    */

    if( !ClaimBenchTargetsSemaphore() )
    {
        return;
    }

    /*
    ** We have the semaphore and we won't let it go.  Just exit the app!
    */

    DestroyWindow( ghwndDlg );
    ghwndDlg = NULL;
}

//*****************************************************************************
//*****************************************************************************

VOID RefreshTargetList( VOID )
{
    int             i;
    HWND            hwndTargetList;
    BYTE            byHaId;
    BYTE            byMaxHaId;
    BYTE            byTarget;
    BYTE            byMaxTarget;
    BYTE            byInquiryCDB[6];
    BYTE            byInquiryBuf[SIZE_INQUIRYBUF];
    CHAR            szVendor[9];
    CHAR            szProduct[17];
    BOOL            bExec;
    DWORD           dwASPIStatus;
    WPARAM          iTarget = 0;
    PBENCHTARGET    pbt;
    SRB_HAInquiry   srbHAInquiry;
    SRB_GDEVBlock   srbGDEVBlock;

    /*
    ** Get the semaphore for the target list (if we can). This disables both
    ** the TEST button and the list box.
    */

    if( !ClaimBenchTargetsSemaphore() )
    {
        return;
    }
    gbRefreshNeeded = FALSE;

    /*
    ** Empty the current bench target list and the listbox on the screen.  We
    ** can't be here unless all of the targets are idle so it is safe to just
    ** delete the entries.
    */

    hwndTargetList = GetDlgItem( ghwndDlg, IDC_TARGETLIST );
    SendMessage( hwndTargetList, LB_RESETCONTENT, 0, 0 );

    while( gpbtHead )
    {
        pbt = gpbtHead->pbtNext;
        HeapFree( GetProcessHeap(), 0, gpbtHead );
        gpbtHead = pbt;
    }

    /*
    ** Scan the SCSI bus for targets we can benchmark.  We first make sure
    ** that ASPI is still working with a call to GetASPI32SupportInfo.
    */

    dwASPIStatus = gpfnGetASPI32SupportInfo();
    if( HIBYTE(LOWORD(dwASPIStatus)) == SS_COMP )
    {
        /*
        ** Use support info for host adapter count and loop over all of them.
        */

        byMaxHaId = LOBYTE(LOWORD(dwASPIStatus));
        for( byHaId = 0; byHaId < byMaxHaId; byHaId++ )
        {
            /*
            ** Do a host adapter inquiry to get max target count.  If the 
            ** target count isn't 8 or 16 then go with a default of 8.
            */

            memset( &srbHAInquiry, 0, sizeof(SRB_HAInquiry) );
            srbHAInquiry.SRB_Cmd = SC_HA_INQUIRY;
            srbHAInquiry.SRB_HaId = byHaId;

            gpfnSendASPI32Command( (LPSRB)&srbHAInquiry );
            if( srbHAInquiry.SRB_Status != SS_COMP )
            {
                continue;
            }

            byMaxTarget = srbHAInquiry.HA_Unique[3];
            if( byMaxTarget != 8 && byMaxTarget != 16 )
            {
                byMaxTarget = 8;
            }

            /*
            ** Loop over all the targets on this host adapter.
            */

            for( byTarget = 0; byTarget < byMaxTarget; byTarget++ )
            {
                /*
                ** Issue get device type call to see if there is a device we're
                ** interested in at this address.  We're interested in disk
                ** drives and CDROMs.
                */

                memset( &srbGDEVBlock, 0, sizeof(SRB_GDEVBlock) );
                srbGDEVBlock.SRB_Cmd = SC_GET_DEV_TYPE;
                srbGDEVBlock.SRB_HaId = byHaId;
                srbGDEVBlock.SRB_Target = byTarget;

                gpfnSendASPI32Command( (LPSRB)&srbGDEVBlock );
                if( srbGDEVBlock.SRB_Status != SS_COMP ||
                    (srbGDEVBlock.SRB_DeviceType != DTYPE_DASD &&
                     srbGDEVBlock.SRB_DeviceType != DTYPE_CDROM) )
                {
                    continue;
                }

                /*
                ** Allocate a bench target structure from the heap and start
                ** filling it in.
                */

                pbt = HeapAlloc( GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(BENCHTARGET) );
                if( !pbt )
                {
                    continue;
                }

                pbt->byHaId = byHaId;
                pbt->byTarget = byTarget;
                memset( pbt->dwThroughput, 0xFF, sizeof(DWORD) * SAMPLES );

                /*
                ** Determine the max transfer count of this target.  It will
                ** be the min of the host adapters min count and the size
                ** of our global transfer buffer.
                */

                pbt->dwMaxTransferBytes = *(PDWORD)&srbHAInquiry.HA_Unique[4];
                if( pbt->dwMaxTransferBytes > gab.AB_BufLen )
                {
                    pbt->dwMaxTransferBytes = gab.AB_BufLen;
                }
                if( pbt->dwMaxTransferBytes < 65536lu )
                {
                    pbt->dwMaxTransferBytes = 65536lu;
                }

                /*
                ** Issue an INQUIRY, do it with retry even though unit
                ** attentions shouldn't bother an INQUIRY.
                */

                memset( byInquiryCDB, 0, 6 );
                byInquiryCDB[0] = SCSI_INQUIRY;
                byInquiryCDB[4] = SIZE_INQUIRYBUF;

                bExec = SyncExecWithRetry
                (
                    pbt,
                    SRB_DIR_IN,
                    SIZE_INQUIRYBUF,
                    byInquiryBuf,
                    6,
                    byInquiryCDB
                );

                /*
                ** Make sure the inquiry worked.  If it failed, or if the
                ** inquiry data returns a different device type than we got
                ** before (guards against certain device drivers and against
                ** vendor unique devices).
                */

                if( !bExec || (byInquiryBuf[0] != DTYPE_DASD && byInquiryBuf[0] != DTYPE_CDROM) )
                {
                    HeapFree( GetProcessHeap(), 0, pbt );
                    continue;
                }

                /*
                ** Add this target to the target list and the listbox.
                */

                pbt->pbtNext = gpbtHead;
                gpbtHead = pbt;

                memcpy( szVendor, byInquiryBuf + 8, 8 );
                for( i = 8; i--; )
                {
                    if( szVendor[i] != '\0' && szVendor[i] != ' ' )
                    {
                        break;
                    }
                    szVendor[i] = '\0';
                }
                szVendor[8] = '\0';

                memcpy( szProduct, byInquiryBuf + 16, 16 );
                for( i = 16; i--; )
                {
                    if( szProduct[i] != '\0' && szProduct[i] != ' ' )
                    {
                        break;
                    }
                    szProduct[i] = '\0';
                }
                szProduct[16] = 0;

                wsprintf
                (
                    gszMsg,
                    "[%ld:%ld] %s %s",
                    (DWORD)byHaId,
                    (DWORD)byTarget,
                    szVendor,
                    szProduct
                );

                SendMessage( hwndTargetList, LB_INSERTSTRING, (WPARAM)iTarget, (LPARAM)gszMsg );
                SendMessage( hwndTargetList, LB_SETITEMDATA, (WPARAM)iTarget, (LPARAM)pbt );
                iTarget++;
            }
        }

        if( iTarget )
        {
            SendMessage( hwndTargetList, LB_SETCURSEL, (WPARAM)0, (LPARAM)0 );
            OnCommandTargetList( (WPARAM)MAKELONG( 0, LBN_SELCHANGE ), 0 );
        }
    }

    /*
    ** Release the bench targets semaphore and enable the dialog controls.
    */

    ReleaseBenchTargetsSemaphore();
}

//*****************************************************************************
//*****************************************************************************

DWORD BenchmarkThread( PBENCHTARGET pbt )
{
    int     i;
    BYTE    byCDB[10];
    BYTE    byCapacity[8];
    BOOL    bExec;
    WORD    wSectors;
    DWORD   dwMaxSector;
    DWORD   dwSectorSize;
    DWORD   dwStartSector;
    DWORD   dwTotalSectors;
    DWORD   dwStartTime;
    DWORD   dwStopTime;

    /*
    ** Initialize the display (zero the data points).
    */

    memset( pbt->dwThroughput, 0xFF, sizeof(DWORD)*SAMPLES );
    OnCommandTargetList( (WPARAM)MAKELONG( 0, LBN_SELCHANGE ), 0 );

    /*
    ** Do a READ CAPACITY on the benchmark target.  This will give us the
    ** sector size and maximum number of sectors on the device.
    */

    memset( byCDB, 0, 10 );
    byCDB[0] = SCSI_RD_CAPAC;

    bExec = SyncExecWithRetry( pbt, SRB_DIR_IN, 8, byCapacity, 10, byCDB );
    if( !bExec )
    {
        ReleaseBenchTargetsSemaphore();
        return 0;
    }

    MOVESCSIDWORD( &byCapacity[0], &dwMaxSector );
    MOVESCSIDWORD( &byCapacity[4], &dwSectorSize );

    /*
    ** Loop through each of the samples doing a sequential test for each.
    ** Note that we only do the samples which fit in the transfer buffer.
    */

    for( i = 0; i < SAMPLES && !(((4lu << i) * 1024lu) > pbt->dwMaxTransferBytes); i++ )
    {
        /*
        ** Read as much as we can in two seconds (sequentially).  Note that
        ** we are not using a very precise method of timing these transfers.
        */

        dwStartSector = 0;
        dwTotalSectors = 0;
        wSectors = (WORD)(((4lu << i) * 1024lu) / dwSectorSize);

        dwStartTime = GetTickCount();
        while( ((dwStopTime = GetTickCount()) - dwStartTime) < 3000 )
        {
            /*
            ** Actually read a block waiting for completion.  If the read
            ** fails then just summarily quit with whatever statistics we've
            ** gathered so far.
            */

            memset( byCDB, 0, 10 );
            byCDB[0] = SCSI_READ10;
            MOVESCSIDWORD( &dwStartSector, &byCDB[2] );
            MOVESCSIWORD( &wSectors, &byCDB[7] );

            bExec = SyncExecWithRetry
            (
                pbt,
                SRB_DIR_IN,
                wSectors * dwSectorSize,
                gab.AB_BufPointer,
                10,
                byCDB
            );
            if( !bExec )
            {
                ReleaseBenchTargetsSemaphore();
                return 0;
            }

            /*
            ** Adjust the starting sector.  Wrap if the next transfer will
            ** exceed the limits of the media (highly unlikely in 2 seconds).
            */
                        
            dwStartSector += wSectors;
            dwTotalSectors += wSectors;
            if( dwStartSector + wSectors > dwMaxSector )
            {
                dwStartSector = 0;
            }
        }

        /*
        ** Update the targets throughput display and the progress bar.
        */

        pbt->dwThroughput[i] = dwTotalSectors * dwSectorSize / 1024lu * 1000lu;
        pbt->dwThroughput[i] /= dwStopTime - dwStartTime;
                        
        OnCommandTargetList( (WPARAM)MAKELONG( 0, LBN_SELCHANGE ), 0 );
        SendDlgItemMessage
        (
            ghwndDlg,
            IDC_PROGRESS,
            PBM_SETPOS,
            (WPARAM)(((i+1)*100)/SAMPLES),
            0
        );
    }

    /*
    ** Done!  Clear progress bar for clean display and release semaphore.
    */

    SendDlgItemMessage( ghwndDlg, IDC_PROGRESS, PBM_SETPOS, 0, 0 );
    ReleaseBenchTargetsSemaphore();
    return 0;
}

//*****************************************************************************
//*****************************************************************************

BOOL SyncExecWithRetry
(
    PBENCHTARGET    pbt,
    BYTE            byFlags,
    DWORD           dwBufferBytes,
    PBYTE           pbyBuffer,
    BYTE            byCDBBytes,
    PBYTE           pbyCDB
)
{
    BOOL            bRetry = TRUE;
    DWORD           dwASPIStatus;
    HANDLE          heventExec;
    SRB_ExecSCSICmd srbExec;

    /*
    ** Setup the SRB for this command.
    */

    memset( &srbExec, 0, sizeof(SRB_ExecSCSICmd) );
    memcpy( srbExec.CDBByte, pbyCDB, byCDBBytes );

    srbExec.SRB_Cmd = SC_EXEC_SCSI_CMD;
    srbExec.SRB_HaId = pbt->byHaId;
    srbExec.SRB_Flags = byFlags;
    srbExec.SRB_Target = pbt->byTarget;
    srbExec.SRB_BufLen = dwBufferBytes;
    srbExec.SRB_BufPointer = pbyBuffer;
    srbExec.SRB_SenseLen = SENSE_LEN;
    srbExec.SRB_CDBLen = byCDBBytes;

    /*
    ** Create an event (if possible) and issue the command.  After sending
    ** the command wait for completion.
    */

    RetryExec:
    {
        heventExec = CreateEvent( NULL, TRUE, FALSE, NULL );
        if( heventExec )
        {
            srbExec.SRB_Flags |= SRB_EVENT_NOTIFY;
            srbExec.SRB_PostProc = (LPVOID)heventExec;
        
            dwASPIStatus = gpfnSendASPI32Command( (LPSRB)&srbExec );
            if( dwASPIStatus == SS_PENDING )
            {
                WaitForSingleObject( heventExec, INFINITE );
            }
            CloseHandle( heventExec );
        }
        else
        {
            gpfnSendASPI32Command( (LPSRB)&srbExec );
            while( srbExec.SRB_Status == SS_PENDING );
        }
    }

    /*
    ** Check for errors.  We'll retry on the first unit attention.  Anything
    ** else will generate an error dialog.
    */

    if( srbExec.SRB_Status != SS_COMP )
    {
        if( bRetry && 
            (srbExec.SRB_TargStat != STATUS_CHKCOND ||
            (srbExec.SenseArea[2] & 0x0F) != KEY_UNITATT) )
        {
            bRetry = FALSE;
            goto RetryExec;
        }
        else
        {
            DialogBoxParam
            (
                ghinstSeqTest, 
                MAKEINTRESOURCE(SCSI_FAILURE),
                ghwndDlg,
                SCSIFailureDlgProc,
                (LPARAM)&srbExec
            );
            return FALSE;
        }
    }

    return TRUE;    
}

//*****************************************************************************
//*****************************************************************************

BOOL ClaimBenchTargetsSemaphore( VOID )
{
    DWORD   dwWait;

    /*
    ** Claim the semaphore guarding the target list.  If we can't find the
    ** semaphore, create it.  If we can't claim the semaphore or it won't
    ** allow itself to be created then just bug out.
    */

    if( ghsemBenchTargets )
    {
        dwWait = WaitForSingleObject( ghsemBenchTargets, 0 );
    }
    else
    {
        ghsemBenchTargets = CreateSemaphore( NULL, 0, 1, NULL );
        dwWait = WAIT_OBJECT_0;
    }

    if( !ghsemBenchTargets || dwWait != WAIT_OBJECT_0 )
    {
        return FALSE;
    }

    /*
    ** Disable the TEST button and the target list box.
    */

    EnableWindow( GetDlgItem( ghwndDlg, IDC_TARGETLIST ), FALSE );
    EnableWindow( GetDlgItem( ghwndDlg, IDC_TEST ), FALSE );
    EnableWindow( GetDlgItem( ghwndDlg, IDC_RESCAN ), FALSE );
    EnableWindow( GetDlgItem( ghwndDlg, IDOK ), FALSE );

    return TRUE;
}

//*****************************************************************************
//*****************************************************************************

VOID ReleaseBenchTargetsSemaphore( VOID )
{
    if( ghsemBenchTargets )
    {
        ReleaseSemaphore( ghsemBenchTargets, 1, NULL );
        EnableWindow( GetDlgItem( ghwndDlg, IDC_TARGETLIST ), TRUE );
        EnableWindow( GetDlgItem( ghwndDlg, IDC_TEST ), TRUE );
        EnableWindow( GetDlgItem( ghwndDlg, IDC_RESCAN ), TRUE );
        EnableWindow( GetDlgItem( ghwndDlg, IDOK ), TRUE );
    }
}

//*****************************************************************************
//*****************************************************************************

VOID DisplayWNASPI32Failure( DWORD dwLastError, BYTE byASPIStatus )
{
    char    szFailureMsg[256];
    LPSTR   lpszMsg;

    if( dwLastError == ERROR_DLL_NOT_FOUND )
    {
        LoadString( ghinstSeqTest, IDS_NEED_WNASPI32, szFailureMsg, sizeof(szFailureMsg) );
    }
    else
    {
        lpszMsg = GetString( IDS_NO_WNASPI32 );
        wsprintf( szFailureMsg, lpszMsg, dwLastError, (DWORD)byASPIStatus );
    }
    MessageBox ( NULL, szFailureMsg, gszFullAppName, MB_ICONSTOP );
}

//*****************************************************************************
//*****************************************************************************

BOOL CALLBACK SCSIFailureDlgProc
(
    HWND    hwndDlg,
    UINT    uMsg,
    WPARAM  wParam,
    LPARAM  lParam
)
{
    int                 j;
    PSRB_ExecSCSICmd    psrb;
    
    switch( uMsg )
    {
        case WM_INITDIALOG:
        {
            psrb  = (PSRB_ExecSCSICmd)lParam;
            strcpy( gszMsg2, GetString( IDS_NA ) );

            wsprintf( gszMsg, "%d", (DWORD)psrb->SRB_Target );
            SetWindowText( GetDlgItem( hwndDlg, SF_TARGET_ID ), gszMsg );

            for( j = 0; j < psrb->SRB_CDBLen; j++ )
            {
                wsprintf( &gszMsg[j*3], "%02X ", (DWORD)psrb->CDBByte[j] );
            }
            SetWindowText( GetDlgItem( hwndDlg, SF_CDB_SENT ), gszMsg );

            wsprintf( gszMsg, "%d", (DWORD)psrb->SRB_HaStat );
            SetWindowText( GetDlgItem( hwndDlg, SF_HA_STATUS ), gszMsg );

            wsprintf( gszMsg, "%d", (DWORD)psrb->SRB_TargStat );
            SetWindowText( GetDlgItem( hwndDlg, SF_TARG_STATUS ), gszMsg );

            if( psrb->SRB_TargStat == STATUS_CHKCOND )
            {
                wsprintf( gszMsg, "%d", (DWORD)(psrb->SenseArea[2] & 0x0F) );
                SetWindowText( GetDlgItem( hwndDlg, SF_SENSE_KEY ), gszMsg );

                wsprintf( gszMsg, "%d", (DWORD)psrb->SenseArea[12] );
                SetWindowText( GetDlgItem( hwndDlg, SF_SENSE_CODE ), gszMsg );

                wsprintf( gszMsg, "%d", (DWORD)psrb->SenseArea[13] );
                SetWindowText( GetDlgItem( hwndDlg, SF_SENSE_QUAL ), gszMsg );
            }
            else
            {
                SetWindowText( GetDlgItem( hwndDlg, SF_SENSE_KEY ), gszMsg2 );
                SetWindowText( GetDlgItem( hwndDlg, SF_SENSE_CODE ), gszMsg2 );
                SetWindowText( GetDlgItem( hwndDlg, SF_SENSE_QUAL ), gszMsg2 );
            }

            return TRUE;
        }

        case WM_COMMAND:
        {
            if( LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL )
            {
                EndDialog( hwndDlg, 0 );
                return TRUE;
            }
        }
    }

    return FALSE;
}

//*****************************************************************************
//*****************************************************************************

BOOL CALLBACK AboutDlgProc
(
    HWND    hwndDlg,
    UINT    uMsg,
    WPARAM  wParam,
    LPARAM  lParam
)
{
    if( (uMsg == WM_COMMAND) && (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) )
    {
        EndDialog( hwndDlg, 0 );
        return TRUE;
    }
    return FALSE;
}

//*****************************************************************************
//*****************************************************************************

LPSTR GetString( UINT uID )
{
    static char szStringBuf[256];
    static UINT uIDPrev = (UINT)-1;

    if( uID != uIDPrev )
    {                                   
        LoadString( ghinstSeqTest, uID, szStringBuf, 255 );
        uIDPrev = uID;
    }
    return szStringBuf;
}
